#define IDD_BENCH         7600
#define IDD_BENCH_2      17600
#define IDD_BENCH_TOTAL 7699

#define IDE_BENCH2_EDIT            100

#define IDC_BENCH_DICTIONARY       101
#define IDT_BENCH_MEMORY_VAL       102
#define IDC_BENCH_NUM_THREADS      103
#define IDT_BENCH_HARDWARE_THREADS 104

#define IDT_BENCH_VER              105
#define IDT_BENCH_CPU              106
#define IDT_BENCH_SYS1             107
#define IDT_BENCH_SYS2             108
#define IDT_BENCH_CPU_FEATURE      109

#define IDT_BENCH_COMPRESS_SPEED1  110
#define IDT_BENCH_COMPRESS_SPEED2  111
#define IDT_BENCH_COMPRESS_RATING1 112
#define IDT_BENCH_COMPRESS_RATING2 113
#define IDT_BENCH_COMPRESS_USAGE1  114
#define IDT_BENCH_COMPRESS_USAGE2  115
#define IDT_BENCH_COMPRESS_RPU1    116
#define IDT_BENCH_COMPRESS_RPU2    117

#define IDT_BENCH_DECOMPR_SPEED1   118
#define IDT_BENCH_DECOMPR_SPEED2   119
#define IDT_BENCH_DECOMPR_RATING1  120
#define IDT_BENCH_DECOMPR_RATING2  121
#define IDT_BENCH_DECOMPR_USAGE1   122
#define IDT_BENCH_DECOMPR_USAGE2   123
#define IDT_BENCH_DECOMPR_RPU1     124
#define IDT_BENCH_DECOMPR_RPU2     125

#define IDT_BENCH_TOTAL_RATING_VAL 130
#define IDT_BENCH_TOTAL_RPU_VAL    131
#define IDT_BENCH_TOTAL_USAGE_VAL  133

#define IDT_BENCH_ELAPSED_VAL      140
// #define IDT_BENCH_SIZE_VAL         141
#define IDT_BENCH_PASSES_VAL       142
#define IDC_BENCH_NUM_PASSES       143

#define IDT_BENCH_LOG           160
#define IDT_BENCH_ERROR_MESSAGE 161

#define IDT_BENCH_COMPRESS_SIZE1  170
#define IDT_BENCH_COMPRESS_SIZE2  171
#define IDT_BENCH_DECOMPR_SIZE1   172
#define IDT_BENCH_DECOMPR_SIZE2   173

// #define IDT_BENCH_FREQ_CUR         150
// #define IDT_BENCH_FREQ_RES         151

#define IDB_STOP                 442
#define IDB_RESTART              443

#define IDT_BENCH_DICTIONARY    4006
#define IDT_BENCH_NUM_THREADS   4009

#define IDT_BENCH_SIZE          1007
#define IDT_BENCH_ELAPSED       3900
#define IDT_BENCH_SPEED         3903

#define IDT_BENCH_MEMORY        7601

#define IDG_BENCH_COMPRESSING   7602
#define IDG_BENCH_DECOMPRESSING 7603
#define IDG_BENCH_TOTAL_RATING  7605

#define IDT_BENCH_RATING_LABEL  7604
#define IDT_BENCH_CURRENT       7606
#define IDT_BENCH_RESULTING     7607
#define IDT_BENCH_USAGE_LABEL   7608
#define IDT_BENCH_RPU_LABEL     7609
#define IDT_BENCH_PASSES        7610
#define IDT_BENCH_CURRENT2     (7606+50)
#define IDT_BENCH_RESULTING2   (7607+50)
