#define IDI_ICON  1

#define IDS_MESSAGE_NO_ERRORS           3001

#define IDS_PROGRESS_TESTING            3302
#define IDS_OPENNING                    3303
#define IDS_SCANNING                    3304

#define IDS_MOVING                      6005

#define IDS_CHECKSUM_CALCULATING        7500
#define IDS_CHECKSUM_INFORMATION        7501
#define IDS_CHECKSUM_CRC_DATA           7502
#define IDS_CHECKSUM_CRC_DATA_NAMES     7503
#define IDS_CHECKSUM_CRC_STREAMS_NAMES  7504

#define IDS_INCORRECT_VOLUME_SIZE       7307

#define IDS_MEM_REQUIRES_BIG_MEM        7811
#define IDS_MEM_REQUIRED_MEM_SIZE       7812
#define IDS_MEM_CURRENT_MEM_LIMIT       7813
#define IDS_MEM_USAGE_LIMIT_SET_BY_7ZIP 7814
#define IDS_MEM_RAM_SIZE                7815

#define IDS_MSG_ARC_UNPACKING_WAS_SKIPPED 7822
