/* Sort.h -- Sort functions
: Igor Pavlov : Public domain */

#ifndef ZIP7_INC_SORT_H
#define ZIP7_INC_SORT_H

#include "7zTypes.h"

EXTERN_C_BEGIN

void Z7_FASTCALL HeapSort(UInt32 *p, size_t size);

EXTERN_C_END

#endif
