#include "ProgressDialogRes.h"
#include "../../GuiCommon.rc"

#define xc 172
#define yc 44

IDD_PROGRESS  DIALOG  0, 0, xs, ys  MY_MODAL_DIALOG_STYLE  MY_FONT
CAPTION "Progress"
BEGIN
  PUSHBUTTON  "Cancel", IDCANCEL, bx, by, bxs, bys
  CONTROL     "Progress1", IDC_PROGRESS1, "msctls_progress32", PBS_SMOOTH | WS_BORDER, m, m, xc, 14
END
