#include "ExtractDialog.rc"
#include "CompressDialog.rc"
#include "BenchmarkDialog.rc"
#include "resource2.h"

STRINGTABLE
BEGIN
    IDS_PROGRESS_COMPRESSING  "Compressing"
    IDS_ARCHIVES_COLON  "Archives:"
END
