EXPORTS
  CreateObject PRIVATE

  GetHandlerProperty PRIVATE
  GetNumberOfFormats PRIVATE
  GetHandlerProperty2 PRIVATE
  GetIsArc PRIVATE

  GetNumberOfMethods PRIVATE
  GetMethodProperty PRIVATE
  CreateDecoder PRIVATE
  CreateEncoder PRIVATE

  GetHashers PRIVATE

  SetCodecs PRIVATE

  SetLargePageMode PRIVATE
  SetCaseSensitive PRIVATE

  GetModuleProp PRIVATE
