// RegistryPlugins.h

#ifndef ZIP7_INC_REGISTRY_PLUGINS_H
#define ZIP7_INC_REGISTRY_PLUGINS_H

#include "../../../Common/MyString.h"

/*
enum EPluginType
{
  kPluginTypeFF = 0
};

struct CPluginInfo
{
  EPluginType Type;
  // bool ClassID_Defined;
  // bool OptionsClassID_Defined;
  // FString FilePath;
  // UString Name;
  // CLSID ClassID;
  // CLSID OptionsClassID;
};

// void ReadPluginInfoList(CObjectVector<CPluginInfo> &plugins);
// void ReadFileFolderPluginInfoList(CObjectVector<CPluginInfo> &plugins);
*/

#endif
